/**************************************************************************
*  Copyright (c) 2014 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  08.02.2014  mifi  First Version.
*  07.02.2015  mifi  Rework to use lwip.
**************************************************************************/
#define __IP_DHCP_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include "tal.h"
#include "ipstack.h"
#include "terminal.h"

#include "lwip\prot\dhcp.h"
#include "lwip\netifapi.h"

/*=======================================================================*/
/*  All extern definitions                                               */
/*=======================================================================*/

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static int DhcpInUse = 0;

static ip_dhcp_callback_t DhcpCallback = NULL;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  _IPDhcpCallbackBound                                                 */
/*                                                                       */
/*  Callback is called whenever the DHCP is bound.                       */
/*                                                                       */
/*  In    : netif                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void _IPDhcpCallbackBound (void)
{
   TAL_DEBUG(TAL_DBG_DHCP, "DHCP is bound");
      
   if (DhcpCallback != NULL)
   {
      DhcpCallback();
   }

} /* _IPDhcpCallbackBound */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  IPDhcpStart                                                          */
/*                                                                       */
/*  Start DHCP negotiation for the network interface.                    */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void IPDhcpStart (void)
{
   struct netif *netif;
   
   if (0 == DhcpInUse)
   {
      netif = netif_find("en0");
      if (netif != NULL)
      {
         DhcpInUse = 1;
         netifapi_dhcp_start(netif);
      }   
   }   
} /* IPDhcpStart */

/*************************************************************************/
/*  IPDhcpStop                                                           */
/*                                                                       */
/*  Stop the DHCP client of the network interface.                       */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void IPDhcpStop (void)
{
   struct netif *netif;
   
   if (1 == DhcpInUse)
   {
      netif = netif_find("en0");
      if (netif != NULL)
      {
         DhcpInUse = 0;
         DhcpCallback = NULL;
         netifapi_dhcp_stop(netif);
      }   
   }
} /* IPDhcpStop */

/*************************************************************************/
/*  IPDhcpIsStarted                                                      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: 0 not started / 1 otherwise                                  */
/*************************************************************************/
int IPDhcpIsStarted (void)
{
   return(DhcpInUse);
} /* IPDhcpIsStarted */

/*************************************************************************/
/*  IPDhcpServerGet                                                      */
/*                                                                       */
/*  Get the DHCP server address of the network interface in host order.  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Address in host order                                        */
/*************************************************************************/
uint32_t IPDhcpServerGet (void)
{
   struct netif *netif;
   struct dhcp  *dhcp;

   uint32_t      addr = 0;
   
   if (1 == DhcpInUse)
   {
      netif = netif_find("en0");
      if (netif != NULL)
      {
         dhcp = netif_dhcp_data(netif);
         addr = dhcp->server_ip_addr.addr;
      }
   }      

   return( ntohl(addr) );
} /* IPDhcpServerGet */

/*************************************************************************/
/*  IPDhcpCallbackSet                                                    */
/*                                                                       */
/*  In    : Callback                                                     */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void IPDhcpCallbackSet (ip_dhcp_callback_t callback)
{
   DhcpCallback = callback;
   
} /* IPDhcpCallbackSet */

/*** EOF ***/
